from zrb.helper.typing import (
    Any, Callable, Iterable, List, Mapping, Optional, Union
)
from zrb.helper.typecheck import typechecked
from zrb.helper.accessories.name import get_random_name
from zrb.task.base_task.base_task import BaseTask
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnTriggered, OnWaiting, OnSkipped, OnStarted, OnReady, OnRetry, OnFailed
)
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

import asyncio
import copy


class RunConfig():
    def __init__(
        self,
        fn: Callable[..., Any],
        args: List[Any],
        kwargs: Mapping[Any, Any],
        execution_id: str
    ):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.execution_id = execution_id

    async def run(self):
        return await self.fn(*self.args, **self.kwargs)


@typechecked
class RecurringTask(BaseTask):
    '''
    A class representing a recurring task that is triggered based on
    specified conditions.

    Examples:

        >>> from zrb import RecurringTask
    '''

    def __init__(
        self,
        name: str,
        task: AnyTask,
        triggers: Iterable[AnyTask] = [],
        single_execution: bool = False,
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        upstreams: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checkers: Iterable[AnyTask] = [],
        checking_interval: float = 0,
        retry: int = 0,
        retry_interval: float = 1,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
        return_upstream_result: bool = False
    ):
        self._task: AnyTask = task.copy()
        inputs = list(inputs) + self._task._get_combined_inputs()
        envs = list(envs) + self._task._get_envs()
        env_files = list(env_files) + self._task._get_env_files()
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            should_execute=should_execute,
            return_upstream_result=return_upstream_result,
        )
        self._triggers: List[AnyTask] = [
            trigger.copy() for trigger in triggers
        ]
        self._run_configs: List[RunConfig] = []
        self._single_execution = single_execution

    async def _set_keyval(self, kwargs: Mapping[str, Any], env_prefix: str):
        await super()._set_keyval(kwargs=kwargs, env_prefix=env_prefix)
        new_kwargs = copy.deepcopy(kwargs)
        new_kwargs.update(self.get_input_map())
        trigger_coroutines = []
        for trigger in self._triggers:
            trigger.add_input(*self._get_inputs())
            trigger.add_env(*self._get_envs())
            trigger_coroutines.append(asyncio.create_task(
                trigger._set_keyval(
                    kwargs=new_kwargs, env_prefix=env_prefix
                )
            ))
        await asyncio.gather(*trigger_coroutines)

    async def run(self, *args: Any, **kwargs: Any):
        await asyncio.gather(
            asyncio.create_task(self.__check_trigger(*args, **kwargs)),
            asyncio.create_task(self.__run_from_queue())
        )

    async def __check_trigger(self, *args: Any, **kwargs: Any):
        task_kwargs = {
            key: kwargs[key]
            for key in kwargs if key not in ['_task']
        }
        is_first_time = True
        while True:
            execution_id = get_random_name(
                add_random_digit=True, digit_count=5
            )
            # Create trigger functions
            trigger_functions = []
            for trigger in self._triggers:
                trigger_copy = trigger.copy()
                trigger_copy._set_execution_id(execution_id)
                trigger_function = trigger_copy.to_function(
                    is_async=True, raise_error=False, show_done_info=False
                )
                trigger_functions.append(asyncio.create_task(
                    trigger_function(*args, **task_kwargs)
                ))
            self.print_out_dark('Waiting for next trigger')
            # Mark task as done since trigger has been defined.
            if is_first_time:
                await self._mark_done()
                is_first_time = False
            # Wait for the first task to complete
            _, pending = await asyncio.wait(
                trigger_functions, return_when=asyncio.FIRST_COMPLETED
            )
            # Cancel the remaining tasks
            for process in pending:
                try:
                    process.cancel()
                except Exception as e:
                    self.print_err(e)
            # Run the task
            task_copy = self._task.copy()
            task_copy._set_execution_id(execution_id)
            fn = task_copy.to_function(
                is_async=True, raise_error=False, show_done_info=False
            )
            self.print_out_dark(f'Add execution to the queue: {execution_id}')
            self._run_configs.append(
                RunConfig(
                    fn=fn,
                    args=args,
                    kwargs=task_kwargs,
                    execution_id=execution_id
                )
            )

    async def __run_from_queue(self):
        while True:
            if len(self._run_configs) == 0:
                await asyncio.sleep(0.1)
                continue
            if self._single_execution:
                # Drain the queue, leave only the latest task
                while len(self._run_configs) > 1:
                    run_config = self._run_configs.pop(0)
                    self.print_out_dark(f'Skipping {run_config.execution_id}')
                    self.clear_xcom(execution_id=run_config.execution_id)
            # Run task
            run_config = self._run_configs.pop(0)
            self.print_out_dark(f'Executing {run_config.execution_id}')
            self.print_out_dark(f'{len(self._run_configs)} tasks left')
            await run_config.run()
            self.clear_xcom(execution_id=run_config.execution_id)
            self._play_bell()



from zrb.task.path_watcher import PathWatcher
from zrb.task.cmd_task import CmdTask
import os

CURRENT_DIR = os.path.dirname(__file__)


def test_path_watcher():
    patch_watcher = PathWatcher(
        path=os.path.join(CURRENT_DIR, 'resources', '**', "*.txt"),
        ignored_path=os.path.join(CURRENT_DIR, 'resources', 'a'),
    )
    toucher = CmdTask(
        name='toucher',
        cwd=os.path.join(CURRENT_DIR, 'resources'),
        cmd=[
            'sleep 1',
            'touch a/a.txt',
            'touch b/b.txt',
            'touch c/c.txt',
        ]
    )
    task = CmdTask(
        name='task',
        upstreams=[patch_watcher, toucher],
        cmd='echo {{task.get_xcom("watch-path.file")}}'
    )
    fn = task.to_function()
    result = fn()
    output = result.output
    assert output == os.path.join(CURRENT_DIR, 'resources', 'b', 'b.txt')



from zrb.helper.typecheck import typechecked
from zrb.task.task import Task


@typechecked
def show_lines(task: Task, *lines: str):
    separator = '\n    '
    task.print_out(
        '\n' + separator + separator.join(lines) + '\n', trim_message=False
    )



from zrb import CmdTask, runner
from zrb.builtin.group import project_group
from .container import start_snake_zrb_app_name_container

###############################################################################
# ⚙️ start-kebab-zrb-task-name
###############################################################################

start_snake_zrb_app_name = CmdTask(
    icon='🚤',
    name='start-kebab-zrb-app-name',
    description='Start human readable zrb app name (as container)',
    group=project_group,
    upstreams=[start_snake_zrb_app_name_container]
)
runner.register(start_snake_zrb_app_name)



import inspect
import re


def get_markdown_from_docstring(cls) -> str:
    """
    Convert Google Style docstrings of a class and its methods to Markdown.

    Args:
        cls (class): The class whose docstrings are to be converted.

    Returns:
        str: The converted Markdown text.
    """
    markdown = f"## `{cls.__name__}`\n\n"
    cls_doc = get_doc(cls)
    markdown += parse_docstring(cls_doc) + '\n'
    for method_name, _ in inspect.getmembers(cls, predicate=inspect.isfunction):  # noqa
        if method_name.startswith('__'):
            # Don't parse private or protected function
            continue
        markdown += f"\n### `{cls.__name__}.{method_name}`\n\n"
        method_doc = get_method_doc(cls, method_name)
        markdown += parse_docstring(method_doc) + '\n'
    return markdown


def get_method_doc(cls, method_name: str) -> str:
    if not hasattr(cls, method_name):
        return None
    method = getattr(cls, method_name)
    if not method.__doc__ and hasattr(cls, '__bases__'):
        for parent in cls.__bases__:
            parent_method_doc = get_method_doc(parent, method_name)
            if parent_method_doc:
                return parent_method_doc
    return method.__doc__


def get_doc(cls) -> str:
    if not cls.__doc__ and hasattr(cls, '__bases__'):
        for parent in cls.__bases__:
            parent_doc = get_doc(parent)
            if parent_doc:
                return parent_doc
    return cls.__doc__


def parse_docstring(docstring: str) -> str:
    if not docstring:
        return 'No documentation available.\n'
    # Split the docstring into lines
    lines = docstring.strip().split('\n')
    line_length = len(lines)
    # Process each line
    markdown_lines = []
    section = ''
    is_previous_code = False
    has_unclosed_backtick = False
    for line_index, line in enumerate(lines):
        line = line.strip()
        is_code = line.startswith('>>> ')
        is_last_line = line_index == line_length - 1
        is_new_section = False
        # Add code backticks
        if is_code and not is_previous_code:
            markdown_lines.append('```python')
            has_unclosed_backtick = True
        elif section == 'examples' and is_previous_code and not is_code:
            markdown_lines.append('```')
            markdown_lines.append('')
            markdown_lines.append('```')
            has_unclosed_backtick = True
        elif is_previous_code and not is_code:
            markdown_lines.append('````')
            markdown_lines.append('')
            has_unclosed_backtick = False
        # Handle lines
        if is_code:
            markdown_lines.append(line[4:])
        elif (
            line.startswith('Attributes:') or
            line.startswith('Attribute:') or
            line.startswith('Attrs:') or
            line.startswith('Attr:')
        ):
            markdown_lines.append('__Attributes:__\n')
            section = 'attributes'
            is_new_section = True
        elif (
            line.startswith('Args:') or
            line.startswith('Arg:') or
            line.startswith('Arguments:') or
            line.startswith('Argument:')
        ):
            markdown_lines.append('__Arguments:__\n')
            section = 'args'
            is_new_section = True
        elif line.startswith('Returns:'):
            markdown_lines.append('__Returns:__\n')
            section = 'returns'
            is_new_section = True
        elif (
            line.startswith('Examples:') or
            line.startswith('Examples:')
        ):
            markdown_lines.append('__Examples:__\n')
            section = 'examples'
            is_new_section = True
        elif line == '```':
            markdown_lines.append(line)
        elif section == 'args' or section == 'attributes':
            named_param_match = re.match(r'^(\w+)\s+\((.+)\):(.+)$', line)
            if named_param_match:
                param_name, param_type, param_desc = named_param_match.groups()
                markdown_lines.append(
                    f'- `{param_name}` (`{param_type}`): {param_desc.strip()}'
                )
            else:
                markdown_lines.append(line)
        elif section == 'returns':
            return_match = re.match(r'^(.+):(.+)$', line)
            if return_match:
                param_type, param_desc = return_match.groups()
                markdown_lines.append(f'`{param_type}`: {param_desc.strip()}')
            else:
                markdown_lines.append(line)
        else:
            markdown_lines.append(line)
        is_previous_code = is_code
        if (is_new_section or is_last_line) and has_unclosed_backtick:
            markdown_lines.append('```')
            markdown_lines.append('')
            has_unclosed_backtick = False
    return '\n'.join(markdown_lines)



from typing import Any, Mapping
from zrb import Task, python_task, runner
from zrb.builtin.group import project_group

###############################################################################
# ⚙️ kebab-zrb-task-name
###############################################################################


@python_task(
    name='kebab-zrb-task-name',
    description='human readable zrb task name',
    group=project_group,
    runner=runner
)
async def snake_zrb_task_name(*args: Any, **kwargs: Any) -> Any:
    task: Task = kwargs.get('_task')
    env_map: Mapping[str, str] = task.get_env_map()
    input_map: Mapping[str, str] = task.get_input_map()
    task.print_out(f'Env map: {env_map}')
    task.print_out(f'Input map: {input_map}')
    return 'ok'



from zrb.helper.typing import Any
from zrb.task.decorator import python_task
from zrb.task_input.str_input import StrInput
from zrb.runner import runner

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='eval',
    inputs=[
        StrInput(
            name='expression',
            shortcut='e',
            default='',
            description='Python expression',
        )
    ],
    description='Evaluate Python expression',
    retry=0,
    runner=runner
)
async def evaluate(*args: str, **kwargs: Any):
    expression: str = kwargs.get('expression', '')
    return eval(expression)



from zrb.helper.typing import (
    Any, Callable, Iterable, Optional, Union
)
from zrb.helper.typecheck import typechecked
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnTriggered, OnWaiting, OnSkipped, OnStarted, OnReady, OnRetry, OnFailed
)
from zrb.task_input.any_input import AnyInput
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.action.runner import Runner
from zrb.task.task import Task

# flake8: noqa E501


@typechecked
def python_task(
    name: str,
    group: Optional[Group] = None,
    inputs: Iterable[AnyInput] = [],
    envs: Iterable[Env] = [],
    env_files: Iterable[EnvFile] = [],
    icon: Optional[str] = None,
    color: Optional[str] = None,
    description: str = '',
    upstreams: Iterable[AnyTask] = [],
    on_triggered: Optional[OnTriggered] = None,
    on_waiting: Optional[OnWaiting] = None,
    on_skipped: Optional[OnSkipped] = None,
    on_started: Optional[OnStarted] = None,
    on_ready: Optional[OnReady] = None,
    on_retry: Optional[OnRetry] = None,
    on_failed: Optional[OnFailed] = None,
    checkers: Iterable[AnyTask] = [],
    checking_interval: float = 0.1,
    retry: int = 2,
    retry_interval: float = 1,
    should_execute: Union[bool, str, Callable[..., bool]] = True,
    return_upstream_result: bool = False,
    runner: Optional[Runner] = None
) -> Callable[[Callable[..., Any]], Task]:
    '''
    python_task decorator helps you turn any Python function into a task

    Returns:
        Callable[[Callable[..., Any]], Task]: A callable turning function into task.
    
    Examples:
        >>> from zrb import python_task
        >>> @python_task(
        >>>    name='my-task'
        >>> )
        >>> def my_task(*args, **kwargs):
        >>>    return 'hello world'
        >>> print(my_task)
        <Task name=my-task>
    '''
    def _create_task(fn: Callable[..., Any]) -> Task:
        task = Task(
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            run=fn,
            should_execute=should_execute,
            return_upstream_result=return_upstream_result
        )
        if runner is not None:
            runner.register(task)
        return task
    return _create_task



from zrb import DockerComposeTask, runner
from zrb.builtin.group import project_group
from ._checker import (
    app_container_checker, rabbitmq_checker, rabbitmq_management_checker,
    redpanda_console_checker, kafka_outside_checker, kafka_plaintext_checker,
    pandaproxy_outside_checker, pandaproxy_plaintext_checker
)
from ._input import (
    local_input, run_mode_input, enable_monitoring_input, host_input,
    https_input, image_input
)
from ._env import app_enable_otel_env, image_env
from ._env_file import compose_env_file
from ._helper import (
    activate_selected_compose_profile, activate_all_compose_profile,
    should_start_container
)
from ._config import RESOURCE_DIR, SERVICE_CONFIGS
from .image import build_snake_zrb_app_name_image


###############################################################################
# ⚙️ remove-kebab-zrb-task-name-container
###############################################################################

remove_snake_zrb_app_name_container = DockerComposeTask(
    icon='💨',
    name='remove-kebab-zrb-app-name-container',
    description='Remove human readable zrb app name container',
    group=project_group,
    cwd=RESOURCE_DIR,
    setup_cmd=activate_all_compose_profile,
    compose_cmd='down',
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    compose_service_configs=SERVICE_CONFIGS,
    envs=[image_env],
    env_files=[compose_env_file]
)
runner.register(remove_snake_zrb_app_name_container)

###############################################################################
# ⚙️ init-kebab-zrb-task-name-container
###############################################################################

init_snake_zrb_app_name_container = DockerComposeTask(
    icon='🔥',
    name='init-kebab-zrb-app-name-container',
    group=project_group,
    inputs=[
        local_input,
        enable_monitoring_input,
        run_mode_input,
        host_input,
        image_input,
    ],
    should_execute=should_start_container,
    upstreams=[
        build_snake_zrb_app_name_image,
        remove_snake_zrb_app_name_container
    ],
    cwd=RESOURCE_DIR,
    setup_cmd=activate_selected_compose_profile,
    compose_cmd='up',
    compose_flags=['-d'],
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    compose_service_configs=SERVICE_CONFIGS,
    envs=[
        image_env,
        app_enable_otel_env,
    ],
    env_files=[compose_env_file],
)

###############################################################################
# ⚙️ start-kebab-zrb-task-name-container
###############################################################################

start_snake_zrb_app_name_container = DockerComposeTask(
    icon='🐳',
    name='start-kebab-zrb-app-name-container',
    description='Start human readable zrb app name container',
    group=project_group,
    inputs=[
        local_input,
        run_mode_input,
        host_input,
        https_input,
        image_input,
    ],
    should_execute=should_start_container,
    upstreams=[init_snake_zrb_app_name_container],
    cwd=RESOURCE_DIR,
    setup_cmd=activate_selected_compose_profile,
    compose_cmd='logs',
    compose_flags=['-f'],
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    compose_service_configs=SERVICE_CONFIGS,
    envs=[image_env],
    env_files=[compose_env_file],
    checkers=[
        app_container_checker,
        rabbitmq_checker,
        rabbitmq_management_checker,
        kafka_outside_checker,
        kafka_plaintext_checker,
        redpanda_console_checker,
        pandaproxy_outside_checker,
        pandaproxy_plaintext_checker,
    ]
)
runner.register(start_snake_zrb_app_name_container)

###############################################################################
# ⚙️ stop-kebab-zrb-task-name-container
###############################################################################

stop_snake_zrb_app_name_container = DockerComposeTask(
    icon='⛔',
    name='stop-kebab-zrb-app-name-container',
    description='Stop human readable zrb app name container',
    group=project_group,
    cwd=RESOURCE_DIR,
    setup_cmd=activate_all_compose_profile,
    compose_cmd='stop',
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    compose_service_configs=SERVICE_CONFIGS,
    envs=[image_env],
    env_files=[compose_env_file]
)
runner.register(stop_snake_zrb_app_name_container)



from locust import HttpUser, task, between
import os

access_token_cookie_key = os.getenv(
    'PUBLIC_AUTH_ACCESS_TOKEN_COOKIE_KEY', 'access_token'
)
admin_username = os.getenv('APP_AUTH_ADMIN_USERNAME', 'root')
admin_password = os.getenv('APP_AUTH_ADMIN_PASSWORD', 'toor')


class QuickstartUser(HttpUser):
    wait_time = between(1, 5)

    @task(2)
    def homepage(self):
        self.client.get('/')

    @task
    def permissions(self):
        self.client.get('/api/v1/auth/permissions')

    def on_start(self):
        login_response = self.client.post(
            '/api/v1/auth/login',
            json={
                'identity': admin_username,
                'password': admin_password
            }
        )
        login_response_json = login_response.json()
        access_token = login_response_json.get('access_token')
        self.client.cookies.set(access_token_cookie_key, access_token)
